import json

from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Misc import random_string

from osbot_utils.utils.Json import json_dumps, json_loads
from osbot_aws.apis.SQS import SQS

class SQS_Queue:                                        # todo refactor main methods to SSM class
    def __init__(self, queue_name=None, queue_url=None, fifo_message_group_id=False):
        self.queue_name            = queue_name
        self.queue_url             = queue_url
        self.fifo_message_group_id = fifo_message_group_id or random_string(prefix="message_group_id-")

    @cache_on_self
    def sqs(self):
        return SQS()

    # main methods

    def add(self, body, message_group_id=None, attributes_data=None):
        self.message_send(body=body, message_group_id=message_group_id, attributes_data=attributes_data)
        return self

    def add_fifo(self, body, fifo_message_group_id=None, attributes_data=None):
        self.message_send_fifo(body=body, fifo_message_group_id=fifo_message_group_id, attributes_data=attributes_data)
        return self

    def arn(self):
        return self.attributes().get('QueueArn')

    def attributes(self, attribute_names='All'):
        return self.sqs().queue_attributes(queue_url=self.url(),attribute_names=attribute_names)

    def attributes_update(self, new_attributes):
        return self.sqs().queue_attributes_update(queue_url=self.url(), new_attributes=new_attributes)

    def create(self, attributes=None):
        return self.sqs().queue_create(queue_name=self.queue_name, attributes=attributes)

    def create_fifo(self, attributes=None):
        return self.sqs().queue_create_fifo(queue_name=self.queue_name, attributes=attributes)

    def delete(self):
        return self.sqs().queue_delete(queue_url=self.url())

    def exists(self):
        return self.sqs().queue_exists(queue_url=self.url())

    def get_message(self, delete_message=True):
        return self.sqs().queue_message_get(queue_url=self.url(), delete_message=delete_message)

    def get_message_with_attributes(self, delete_message=True):
        return self.sqs().queue_message_get_with_attributes(queue_url=self.url(), delete_message=delete_message)

    def get_many(self, count=1, delete_message=True):
        return self.sqs().queue_messages_get_n(queue_url=self.url(), n=count, delete_message=delete_message)

    def info(self):                         # consistent method with similar classes like Lambda
        return self.sqs().queue_info(queue_url=self.url())

    def fifo(self):
        return self.info().get('FifoQueue') == 'true'

    def message_raw(self):
        return self.sqs().queue_message_get_raw(queue_url=self.url())

    def message_delete(self, receipt_handle):
        return self.sqs().queue_message_delete(queue_url=self.url(), receipt_handle=receipt_handle)

    def message_send(self, body, message_group_id=None, attributes_data=None):
        return self.sqs().queue_message_send(queue_url=self.url(), body=body, message_group_id=message_group_id, attributes_data=attributes_data)

    def message_send_fifo(self,body, fifo_message_group_id=None, attributes_data=None):
        message_group_id = fifo_message_group_id or self.fifo_message_group_id  # allow this value to be overwritten per request
        return self.message_send(body=body, message_group_id= message_group_id, attributes_data=attributes_data)

    def messages_in_queue(self):
        return self.sqs().queue_messages_count(queue_url=self.url())

    def messages_not_visible(self):
        return self.sqs().queue_messages_count_not_visible(queue_url=self.url())

    def name(self):
        return self.queue_name

    def permission_add(self, label, aws_account_ids, actions):
        return self.sqs().permission_add(queue_url=self.url(), label=label, aws_account_ids=aws_account_ids, actions=actions)

    def permission_add_for_service(self, source_arn, service, resource):
        return self.sqs().permission_add_for_service(queue_url=self.url(), service=service, source_arn=source_arn, resource=resource)

    def permission_delete(self, label):
        return self.sqs().permission_delete(queue_url=self.url(), label=label)

    def permissions(self):
        return self.sqs().permissions(queue_url=self.url())

    def policy(self):
        return self.sqs().policy(queue_url=self.url())

    def push(self, data, message_group_id=None, attributes_data=None):
        if data:
            if type(data) is not str:
                body = json_dumps(data)
            else:
                body = data
            self.message_send(body=body, message_group_id=message_group_id, attributes_data=attributes_data)
        return self

    def push_fifo(self, data, fifo_message_group_id=None, attributes_data=None):
        if data:
            if type(data) is not str:
                body = json_dumps(data)
            else:
                body = data
            self.message_send_fifo(body=body, fifo_message_group_id=fifo_message_group_id,attributes_data=attributes_data)
        return self

    def pop(self, delete_message=True):
        data_json = self.get_message(delete_message=delete_message)
        if data_json:
            return json_loads(data_json)
        return None

    def url(self):
        if self.queue_url  is None:
            self.queue_url = self.sqs().queue_url(queue_name=self.queue_name)
        return self.queue_url

SQS_Queue.next = SQS_Queue.pop
SQS_Queue.pull = SQS_Queue.pop
SQS_Queue.size = SQS_Queue.messages_in_queue


from dotenv import load_dotenv
from osbot_aws.apis.shell.Shell_Client import Shell_Client

from osbot_aws.AWS_Config import AWS_Config

from osbot_utils.testing.Duration import Duration

from osbot_aws.helpers.Lambda_Layer_Create import Lambda_Layer_Create
from osbot_utils.utils.Misc import wait_for

from osbot_aws.OSBot_Setup import OSBot_Setup
from osbot_aws.apis.test_helpers.Temp_Aws_Roles import Temp_Aws_Roles
from osbot_aws.helpers.Lambda_Package import Lambda_Package

class Deploy_Lambda:

    def __init__(self, handler, stage=None):
        load_dotenv()
        self.osbot_setup          = OSBot_Setup()
        self.aws_config           = self.osbot_setup.aws_config
        if type(handler) is str:
            self.handler          = None
            self.module_name      = handler
        else:
            self.handler          = handler
            self.module_name      = handler.__module__

        self.stage                = stage
        self.role_arn             = Temp_Aws_Roles().for_lambda_invocation__role_arn()
        # self.layers               = []
        # self.env_variables        = {}
        self.package              = self.get_package()

    def add_function_source_code(self):
        root_module_name = self.handler.__module__.split(".").pop(0)
        self.package.add_module(root_module_name)

    def add_folder(self, source, ignore=None):
        self.package.add_folder(source=source, ignore=ignore)

    def add_layer(self, layer_arn):
        self.package.add_layer(layer_arn)
        return self

    def add_layers(self, layers_arn):
        for layer_arn in layers_arn:
            self.package.add_layer(layer_arn)
        return self

    def add_module(self, module_name):
        self.package.add_module(module_name)
        return self

    def add_osbot_aws    (self):    self.package.add_osbot_aws    () ; return self
    def add_osbot_browser(self):    self.package.add_osbot_browser() ; return self
    def add_osbot_utils  (self):    self.package.add_osbot_utils  () ; return self
    def add_osbot_elastic(self):    self.package.add_osbot_elastic() ; return self

    def delete(self):
        return self.lambda_function().delete()

    def deploy(self):
        return self.update() == 'Successful'

    def exists(self):
        return self.package.aws_lambda.exists()

    def invoke(self, params=None):
        return self.lambda_function().invoke(params)

    def invoke_async(self, params=None):
        return self.lambda_function().invoke_async(params)

    def invoke_return_logs(self, params=None):
        return self.lambda_function().invoke_return_logs(params)

    def files(self):
        return self.package.get_files()

    def function_url(self):
        return self.lambda_function().function_url()

    def get_package(self):
        package = Lambda_Package(self.lambda_name())
        package.aws_lambda.set_s3_bucket(self.osbot_setup.s3_bucket_lambdas)
        package.aws_lambda.set_role(self.role_arn)
        return package

    def lambda_name(self):
        if self.stage is None:
            return self.module_name
        return f"{self.module_name}-{self.stage}"

    def lambda_function(self):
        return self.package.aws_lambda

    def lambda_shell(self):
        return Shell_Client(self.lambda_function())

    def update(self, wait_for_update=True):
        if self.uses_container_image() is False:
            self.add_function_source_code()
            if len(self.package.get_files()) == 0:                       # todo: add this check to the package.update()  method
                raise Exception("There are not files to deploy")
        update_result= self.package.update()
        if update_result.get('status') == 'ok':
            if wait_for_update:
                return self.lambda_function().wait_for_function_update_to_complete()
        return update_result

    def set_container_image(self, image_uri):
        self.package.set_image_uri(image_uri)

    def set_handler(self, handler):
        self.package.set_handler(handler)
        return self

    def set_layers(self, layers):
        self.package.set_layers(layers)

    def set_packages_using_layer(self, packages, skip_layer_creation_if_exists=True):
        layer_name          = f'layer_for__{self.lambda_name()}'
        lambda_layer_create = Lambda_Layer_Create(layer_name)
        lambda_layer_create.add_packages(packages)
        layer_arn           = lambda_layer_create.create(skip_if_exists=skip_layer_creation_if_exists)
        self.lambda_function().add_layer(layer_arn)
        return lambda_layer_create

    def set_env_variable(self, key, value):
        self.package.set_env_variable(key, value)
        return self

    def set_env_variables(self, env_variables):
        self.package.set_env_variables(env_variables)

    def uses_container_image(self):
        return self.package.uses_image_uri()

# legacy methods
Deploy_Lambda.lambda_invoke = Deploy_Lambda.invoke



from osbot_aws.aws.ec2.EC2 import EC2
from osbot_aws.aws.iam.utils.Temp_Role__For_Service import Temp_Role__For_Service


class EC2__with_temp_role(Temp_Role__For_Service, EC2):

    def __init__(self):
        super().__init__()
        self._temp_role_config.boto3_service_name = 'ec2'
        self._temp_role_config.required_services  = ['ec2']



import boto3
import json
import time

from osbot_aws.apis.Session import Session


class Athena:
    def __init__(self):
        self.aws_athena = Session().client('athena')

    def execute_query_and_return_csv(self, query, database,s3_output):
        query_id = self.start_query(query, database, s3_output)
        self.wait_for_query(query_id)
        return self.query_csv(query_id)

    def query_csv(self,query_id):
        details = self.query_details(query_id)
        return details.get('QueryExecution').get('ResultConfiguration').get('OutputLocation')

    def query_details(self, query_id):
        return  self.aws_athena.get_query_execution(QueryExecutionId=query_id)

    def query_status(self, query_id):
        exectution = self.aws_athena.get_query_execution(QueryExecutionId=query_id)
        return exectution.get('QueryExecution').get('Status').get('State')

    def start_query(self, query, database, s3_output):
        response = self.aws_athena.start_query_execution(
            QueryString=query,
            QueryExecutionContext={
                'Database': database
            },
            ResultConfiguration={
                'OutputLocation': s3_output,
            }
        )
        return response.get('QueryExecutionId')

    def query_as_json(self, query_id):
        return json.dumps(self.query_results(query_id))

    def query_results(self, query_id, next_Token = None):
        if next_Token is None:
            results = self.aws_athena.get_query_results(QueryExecutionId = query_id)
        else:
            results = self.aws_athena.get_query_results(QueryExecutionId=query_id, NextToken=next_Token)

        headers       = [h['Name'] for h in results['ResultSet']['ResultSetMetadata']['ColumnInfo']]
        query_results = []

        for i, row in enumerate(results['ResultSet']['Rows']):
            if i == 0 and next_Token is None:
                continue
            row_data = {}
            for j, value in enumerate(row['Data']):
                row_data[headers[j]] = value['VarCharValue']
            query_results.append(row_data)
        return query_results, results.get('NextToken')

    def wait_for_query(self, query_id):
        while True:
            status = self.query_status(query_id)
            if status in ['SUCCEEDED', 'FAILED', 'CANCELLED']:
                return self
            time.sleep(0.2)     # 200,s






from osbot_utils.utils.Lists import array_get

from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_utils.utils.Json import json_loads, json_to_str

from osbot_utils.utils.Misc import list_set

from osbot_utils.decorators.methods.catch import catch
from osbot_utils.decorators.methods.cache import cache

from osbot_aws.apis.Session import Session


class SQS:

    @cache_on_self
    def client(self):
        return Session().client('sqs')

    @catch
    def queue_attributes(self, queue_url, attribute_names='All'):
        queue_data = self.client().get_queue_attributes(QueueUrl=queue_url, AttributeNames=[attribute_names]).get('Attributes')
        queue_data['QueueUrl' ] = queue_url                              # add this important value (which is missing from the current data returned from AWS)
        queue_data['QueueName'] = queue_data['QueueUrl' ].split('/').pop()
        return queue_data

    def queue_attributes_update(self, queue_url, new_attributes):
        return self.client().set_queue_attributes(QueueUrl=queue_url, Attributes=new_attributes)

    def queue_create(self, queue_name, attributes=None):
        kwargs = { "QueueName": queue_name,
                   "Attributes": attributes or {} }
        return self.client().create_queue(**kwargs).get('QueueUrl')

    def queue_create_fifo(self, queue_name, attributes=None):
        if attributes is None:
            attributes = {}
        attributes.update({'FifoQueue'                : 'True'  ,
                           'ContentBasedDeduplication': 'True' })

        return self.queue_create(queue_name=queue_name, attributes=attributes)

    def queue_delete(self, queue_url):
        if self.queue_exists(queue_url=queue_url) is False: return False
        self.client().delete_queue(QueueUrl=queue_url)
        return self.queue_exists(queue_url=queue_url) is False

    def queue_exists(self, queue_url):
        return self.queue_info(queue_url=queue_url).get('error') is None

    def queue_info(self, queue_url):                         # consistent method with similar classes like Lambda
        return self.queue_attributes(queue_url=queue_url)

    def queue_message_delete(self, queue_url, receipt_handle):
        return self.client().delete_message(QueueUrl=queue_url, ReceiptHandle=receipt_handle)

    def queue_message_get(self, queue_url, delete_message=True):
        message = self.queue_message_get_raw(queue_url)
        if message:
            body = message.get('Body')
            receipt_handle = message.get('ReceiptHandle')
            if delete_message:
                self.queue_message_delete(queue_url, receipt_handle)
            return body

    def queue_message_get_raw(self, queue_url, message_attribute_names=None):
        kwargs = {"QueueUrl"             : queue_url,
                  "MessageAttributeNames": message_attribute_names or ['All'],
                  "MaxNumberOfMessages"  : 1 }
        messages = self.client().receive_message(**kwargs).get('Messages')
        return array_get(messages,0)                        # there will only be a max of one message in queue

    def queue_message_get_with_attributes(self, queue_url, delete_message=True):
        message = self.queue_message_get_raw(queue_url=queue_url)
        if message:
            body            = message.get('Body')
            attributes      = message.get('MessageAttributes')
            receipt_handle  = message.get('ReceiptHandle')
            attributes_data = {}
            if attributes:
                for key,value in attributes.items():
                    attributes_data[key] = value.get('StringValue')
            if delete_message:
                self.queue_message_delete(queue_url=queue_url,receipt_handle=receipt_handle)
            return body, attributes_data

    def queue_messages_get_n(self, queue_url, n=1, delete_message=True):      # todo see if we can use client().receive_message with a higher value of MaxNumberOfMessages
        messages = []
        for i in range(0,n):
            message = self.queue_message_get(queue_url=queue_url, delete_message=delete_message)
            if message:
                messages.append(message)
            else:
                break
        return messages

    def queue_message_send(self, queue_url, body, message_group_id=None, attributes_data=None):
        kwargs = {"QueueUrl"         : queue_url,
                  "MessageBody"      : body     ,
                  "MessageAttributes": {}}

        if attributes_data:
            for key,value in attributes_data.items():
                kwargs['MessageAttributes'][key] = { 'StringValue': value , 'DataType': 'String'}
        if message_group_id:
            kwargs["MessageGroupId"] = message_group_id

        return self.client().send_message(**kwargs).get('MessageId')

    def queue_message_send_fifo(self, queue_url, message_group_id, body, attributes_data=None):
        return self.queue_message_send(queue_url=queue_url, message_group_id=message_group_id, body=body, attributes_data=attributes_data)

    def queue_messages_count(self, queue_url):
        return int(self.queue_attributes(queue_url=queue_url).get('ApproximateNumberOfMessages'))

    def queue_messages_count_not_visible(self, queue_url):
        return int(self.queue_attributes(queue_url=queue_url).get('ApproximateNumberOfMessagesNotVisible'))

    def queue_url(self, queue_name):
        return self.client().get_queue_url(QueueName=queue_name).get('QueueUrl')

    def queue_url_from_queue_arn(self, queue_arn):
        result = self.queues(index_by='QueueArn')
        return result.get(queue_arn, {}).get('QueueUrl')

    @index_by
    def queues(self):                               # todo: see if there is a better way to get this info about existing queues
        """Note: there could be missing entries since the list of queues doesn't update as fast as it should
                 (i.e. SQS.queues_urls(): might have more entries that SQS.queues() )
        """
        data = []
        for queue_url in self.queues_urls():
            queue_info = self.queue_info(queue_url=queue_url)
            if queue_info.get('error') is None:
                data.append(queue_info)
        return data

    def queues_arns(self):
        return list_set(self.queues(index_by="QueueArn"))

    def queues_urls(self):
        return self.client().list_queues().get('QueueUrls')

    @catch
    def permission_add(self, queue_url, label, aws_account_ids, actions):
        self.permission_delete(queue_url, label)  # delete if already exists
        params = {'QueueUrl'     : queue_url       ,
                  'Label'        : label           ,
                  'AWSAccountIds': aws_account_ids ,
                  'Actions'      : actions         }
        return self.client().add_permission(**params)

    # todo: see if there is a better way to do this. It shouldn't be needed to write this method, but boto3's add_permission doesn't seem to have the ability to control the Condition and Pricipal values
    def permission_add_for_service(self, queue_url, source_arn, service, resource, action='sqs:SendMessage', effect='Allow'):
        policy_statement_id = f'{action}-rule-{source_arn}'
        statement  = { 'Action'   : action,
                       'Condition': {'ArnEquals': {'aws:SourceArn':source_arn}},
                       'Effect'   : effect,
                       'Principal': {'Service': service},
                       'Resource' : resource,
                       'Sid'      : policy_statement_id }

        policy = self.policy(queue_url=queue_url)
        if policy == {}:
            policy = {'Statement' : []           ,
                      'Version'   : '2008-10-17' }
        policy.get('Statement').append(statement)
        policy_str = json_to_str(policy)
        self.queue_attributes_update(queue_url=queue_url, new_attributes={"Policy": policy_str})
        return self.policy(queue_url=queue_url)


    @catch
    def permission_delete(self, queue_url, label):
        params = {  'QueueUrl'     : queue_url      ,
                    'Label'        : label          ,}
        return self.client().remove_permission(**params)

    def permissions(self, queue_url):
        return self.policy(queue_url).get('Statement', [])

    def policy(self, queue_url):
        value = self.queue_attributes(queue_url=queue_url).get('Policy')
        if value:
            return json_loads(value)
        return {}



from unittest                                           import TestCase
from osbot_aws.apis.Lambda                              import Lambda
from osbot_aws.aws.s3.S3                                import S3
from osbot_aws.aws.dynamo_db.Dynamo_DB                  import Dynamo_DB
from osbot_aws.aws.dynamo_db.Dynamo_DB__with_temp_role  import Dynamo_DB__with_temp_role
from osbot_aws.aws.lambda_.Lambda__with_temp_role       import Lambda__with_temp_role
from osbot_aws.aws.s3.S3__with_temp_role                import S3__with_temp_role
from osbot_aws.testing.Pytest                           import skip_pytest___aws_pytest_user_name__is_not_set
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_utils.utils.Misc                             import random_text


class TestCase__Lambda(TestCase):
    client__dynamo_db: object
    client__lambda   : object
    client__s3       : object
    type_Dynamo_DB   : type
    type_Lambda      : type
    type_S3          : type
    dynamo_db        : Dynamo_DB
    lambda_          : Lambda
    s3               : S3
    reset_iam_creds : bool = False

    @classmethod
    def setUpClass(cls):
        skip_pytest___aws_pytest_user_name__is_not_set()
        cls.type_Dynamo_DB    = Dynamo_DB__with_temp_role
        cls.type_Lambda       = Lambda__with_temp_role
        cls.type_S3           = S3__with_temp_role
        cls.dynamo_db         = cls.type_Dynamo_DB()
        cls.lambda_           = cls.type_Lambda()
        cls.s3                = cls.type_S3()
        cls.client__dynamo_db = cls.dynamo_db.client
        cls.client__lambda    = cls.lambda_  .client
        cls.client__s3        = cls.s3       .client

        if cls.reset_iam_creds:
            cls.dynamo_db.temp_role__iam_reset_credentials()
            cls.lambda_  .temp_role__iam_reset_credentials()
            cls.s3       .temp_role__iam_reset_credentials()

    @classmethod
    def tearDownClass(cls):
        pass


import boto3

from osbot_aws.apis.Session import Session
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.utils.Json import to_json_str


class API_Gateway_Management_API:

    def __init__(self, endpoint_url):
        self.endpoint_url = endpoint_url
    def __enter__(self                           ): return self
    def __exit__ (self, exc_type, exc_val, exc_tb): pass


    @cache_on_self
    def client(self):
        return boto3.client('apigatewaymanagementapi', endpoint_url=self.endpoint_url)

    @remove_return_value(field_name='ResponseMetadata')
    def connection(self, connection_id):
        return self.client().get_connection(ConnectionId=connection_id)

    def post_to_connection(self, connection_id, data):
        if type(data) is dict:
            data = to_json_str(data)
        return self.client().post_to_connection(ConnectionId=connection_id, Data=data)


from osbot_utils.utils.Misc import list_set

from osbot_utils.utils.Json import json_save_file, json_load_file

from osbot_aws.apis.Lambda_Layer import Lambda_Layer

from osbot_utils.utils.Process  import Process
from osbot_utils.utils.Files import current_temp_folder, path_combine, folder_create, folder_sub_folders, folders_names, \
    folder_exists, folder_delete_recursively, file_exists

from osbot_aws.helpers.Lambda_Layers_Local import Lambda_Layers_Local

class Lambda_Layer_Create:

    FILE_INSTALLED_PACKAGES = '_osbot_installed_packages.json'

    def __init__(self, layer_name):
        self.layer_name          = layer_name
        self.lambda_layer        = Lambda_Layer(self.layer_name)
        self.lambda_layers_local = Lambda_Layers_Local()
        self.target_aws_lambda   = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add_package(self, package):
        if type(package) is dict:
            if 'name' not in package:
                return {'error'    : 'missing name property in package_name dict',
                        'installed': False                             ,         }
            package_name      = package.get('name')
            target_aws_lambda = package.get('target_aws_lambda', True)
        else:
            package_name      = package
            target_aws_lambda = self.target_aws_lambda

        installed_package = self.installed_packages().get(package_name)
        if installed_package:
            return installed_package

        target_folder = self.path_layer_folder()
        args          = ['install']
        if target_aws_lambda:
            args.extend(['--platform','manylinux1_x86_64', '--only-binary=:all:'])
        args.extend(['-t', target_folder, package_name])

        result      = Process.run('pip3', args)
        output      = result.get('stdout')
        error       = result.get('stderr')
        exists      = result.get('status') == 'ok' and  ('ERROR' in error)  # todo: add better detection to see if package was installed ok
        installed   = True
        result      = dict(exists            = exists           ,
                           installed         = installed        ,
                           name              = package_name     ,
                           output            = output           ,
                           error             = error            ,
                           target_aws_lambda = target_aws_lambda)

        self.update_installed_packages(package_name, result)
        return result

    def add_packages(self, packages):
        results = {}
        for package in packages:
            package      = self.add_package(package)
            package_name = package.get('name')
            results[package_name] = package
        return results

    def arn_latest(self):
        return self.lambda_layer.arn_latest()

    def create(self, skip_if_exists=True):
        if skip_if_exists and self.exists():
            return self.arn_latest()
        return self.lambda_layer.create_from_folder_via_s3(self.path_layer_folder())

    def delete_layer(self):
        return self.lambda_layer.delete()

    def delete_local_layer_folder(self):
        return self.layer_folder_delete()

    def exists(self):
        return self.lambda_layer.exists()

    def has_package_installed(self, package_name):
        return package_name in self.installed_packages()


    def installed_packages(self):
        path_installed_packages = self.path_installed_packages()
        return json_load_file(path_installed_packages)
        # todo: add detection of which folders (and maybe even files) were modified on the installation
        # todo: add back the feature below to extract the versions of the packages installed from the .dist file
        #
        # folders = folder_sub_folders(self.path_layer_folder())
        #
        # # Dictionary to store package names and versions
        # package_dict = {}
        #
        # for folder in folders:
        #     # Split path to get package name and version info
        #     parts = folder.split('/')
        #     name = parts[-1]
        #
        #     # Check for package version info
        #     if name.endswith('.dist-info'):
        #         package_name = name.split('-')[0]
        #         version = name.split('-')[1].replace('.dist', '').split('.dist-info')[0]
        #         package_dict[package_name] = version
        #     else:
        #         if name not in package_dict:
        #             package_dict[name] = None

        # return package_dict

    def installed_packages_names(self):
        return list_set(self.installed_packages())

    def installed_packages_reset(self):
        if file_exists(self.path_installed_packages()):
            json_save_file({}, self.path_installed_packages())
        self.delete_local_layer_folder()
        self.layer_folder_create()
        return self

    def layer_folder_create(self):
        return folder_create(self.path_layer_folder())

    def layer_folder_delete(self):
        return folder_delete_recursively(self.path_layer_folder())

    def layer_folder_exists(self):
        return folder_exists(self.path_layer_folder())

    def path_installed_packages(self):
        return path_combine(self.path_layer_folder(), self.FILE_INSTALLED_PACKAGES)

    def path_lambda_dependencies(self):
        return self.lambda_layers_local.path_lambda_dependencies()

    def path_layer_folder(self):
        return path_combine(self.lambda_layers_local.path_lambda_dependencies(), self.layer_name)

    def recreate(self):
        return self.create(skip_if_exists=False)

    def update_installed_packages(self, package_name, data):
        installed_packages               = self.installed_packages()
        installed_packages[package_name] = data
        path_layer_folder                = self.path_layer_folder()
        path_installed_packages          = self.path_installed_packages()
        folder_create(path_layer_folder)                                        # make sure folder_exists
        json_save_file(installed_packages, path_installed_packages, pretty=True)             # save data into path_installed_packages file
        return installed_packages


from botocore.exceptions import ClientError
from osbot_aws.AWS_Config import AWS_Config
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Misc import wait_for
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.decorators.lists.index_by import index_by

from osbot_aws.apis.Session import Session
from osbot_utils.utils.Status import status_warning, status_ok, status_error

from osbot_aws.aws.sts.STS import STS


class Cloud_Watch_Logs():
    def __init__(self):
        self.aws_config = AWS_Config()
        self.account_id  = self.aws_config.aws_session_account_id()
        self.region_name = self.aws_config.aws_session_region_name()

    @cache_on_self
    def client(self):
        return Session().client('logs')

    @cache_on_self
    def sts(self):
        return STS()

    def destinations(self):
        return self.client().describe_destinations().get('destinations')

    def export_tasks(self):
        return self.client().describe_export_tasks().get('exportTasks')

    def log_events(self, log_group_name, log_stream_name):
        if self.log_stream_exists(log_group_name=log_group_name, log_stream_name=log_stream_name):
            return '\n'.join(self.log_events_messages(log_group_name=log_group_name, log_stream_name=log_stream_name))
        return ''

    def log_events_messages(self, log_group_name, log_stream_name):
        messages = []
        events = self.log_events_raw(log_group_name=log_group_name, log_stream_name=log_stream_name).get('events')
        for event in events:
            messages.append(event.get('message'))
        return messages

    @remove_return_value(field_name='ResponseMetadata')
    def log_events_put(self, log_group_name, log_stream_name, log_events, next_sequence_token=None):
        kwargs = {  'logGroupName' : log_group_name,
                    'logStreamName': log_stream_name,
                    'logEvents'    : log_events }

        if next_sequence_token:
            kwargs.update({'sequenceToken': next_sequence_token})

        result = self.client().put_log_events(**kwargs)
        return result.get('nextSequenceToken')

    @remove_return_value(field_name='ResponseMetadata')
    def log_events_raw(self, log_group_name, log_stream_name):
        result = self.client().get_log_events(logGroupName=log_group_name, logStreamName=log_stream_name)
        return result
        # todo add support for these filters
        # startTime = 123,
        # endTime = 123,
        # nextToken = 'string',
        # limit = 123,
        # startFromHead = True | False

    def log_group(self,log_group_name):
        log_group_arn = self.log_group_arn(log_group_name)
        log_groups    = self.log_groups(log_group_prefix=log_group_name)
        for log_group in log_groups:
            if log_group_arn == log_group.get('arn'):
                return log_group

    def log_group_arn(self, log_group_name):
        return f'arn:aws:logs:{self.region_name}:{self.account_id}:log-group:{log_group_name}:*'

    def log_group_create(self, log_group_name, tags=None):
        if self.log_group_exists(log_group_name):
            return status_warning(f'log group already existed: {log_group_name}')
        kwargs = {'logGroupName' : log_group_name}
        if tags:
            kwargs['tags'] = tags
        self.client().create_log_group(**kwargs)
        return status_ok(message=f"log group created ok: {log_group_name}")

    def log_group_delete(self,log_group_name):
        if self.log_group_exists(log_group_name) is False:
            return status_warning(f'log group did not exist: {log_group_name}')
        self.client().delete_log_group(logGroupName= log_group_name)
        return status_ok(message=f"log group deleted ok: {log_group_name}")

    def log_group_exists(self, log_group_name):
        return self.log_group(log_group_name=log_group_name) is not None

    def log_group_not_exists(self, log_group_name):
        return self.log_group_exists(log_group_name) is False

    @index_by
    def log_groups(self, log_group_prefix=None):
        kwargs = {}
        if log_group_prefix:
            kwargs['logGroupNamePrefix'] = log_group_prefix
        return self.client().describe_log_groups(**kwargs).get('logGroups')     # todo add pagination (see how IAM does it)

    def log_stream_create(self, log_group_name, log_stream_name):
        if self.log_group_not_exists(log_group_name=log_group_name):                    # if the log group doesn't exist, create it
            self.log_group_create(log_group_name).get('status')
        try:
            self.client().create_log_stream(logGroupName=log_group_name, logStreamName=log_stream_name)
            return status_ok(message=f"log stream created ok: {log_stream_name}")
        except ClientError as exception:
            if exception.response['Error']['Code'] == 'ResourceAlreadyExistsException':
                return status_warning(message=f"log stream already existed: {log_stream_name}")
            return status_error(message=f'exception', error=exception)

    def log_stream(self, log_group_name, log_stream_name):                                      # todo: map out the performance implications of .log_streams(...) using describe_log_streams is it looks like it has very low limits of 5 TPS per account per region
        log_stream_arn = self.log_stream_arn(log_group_name=log_group_name, log_stream_name=log_stream_name)
        log_streams    = self.log_streams   (log_group_name=log_group_name, log_stream_name_prefix=log_stream_name)
        for log_stream in log_streams:
            if log_stream.get('arn') == log_stream_arn:
                return log_stream

    def log_stream_arn(self, log_group_name, log_stream_name):
        return f'arn:aws:logs:{self.region_name}:{self.account_id}:log-group:{log_group_name}:log-stream:{log_stream_name}'

    def log_stream_delete(self,log_group_name, log_stream_name):
        if self.log_stream_exists(log_group_name, log_stream_name) is False:
            return status_warning(f'log stream did not exist: {log_group_name}')
        self.client().delete_log_stream(logGroupName= log_group_name, logStreamName=log_stream_name)
        return status_ok(message=f"log stream deleted ok: {log_stream_name}")

    def log_stream_exists(self, log_group_name, log_stream_name):
        return self.log_stream(log_group_name=log_group_name, log_stream_name=log_stream_name) is not None

    def log_stream_next_token(self,log_group_name, log_stream_name):
        log_stream = self.log_stream(log_group_name=log_group_name, log_stream_name=log_stream_name)
        if log_stream:
            return log_stream.get('nextForwardToken')

    def log_stream_not_exists(self, log_group_name, log_stream_name):
        return self.log_stream_exists(log_group_name=log_group_name, log_stream_name=log_stream_name) is False

    def log_streams(self, log_group_name, log_stream_name_prefix=None):     # todo: map out the performance implications of using describe_log_streams is it looks like it has very low limits of 5 TPS per account per region
        if self.log_group_not_exists(log_group_name=log_group_name):
            return []
        kwargs = {'logGroupName': log_group_name}
        if log_stream_name_prefix:
            kwargs['logStreamNamePrefix'] = log_stream_name_prefix
        result = self.client().describe_log_streams(**kwargs)
        return result.get('logStreams')

    def metric_filters(self):
        return self.client().describe_metric_filters().get('metricFilters')

    def queries(self):
        return self.client().describe_queries().get('queries')

    def query_definitions(self):
        return self.client().describe_query_definitions().get('queryDefinitions')

    def resource_policies(self):
        return self.client().describe_resource_policies().get('resourcePolicies')

    def subscription_filters(self, log_group_name):
        return self.client().describe_subscription_filters(logGroupName=log_group_name).get('subscriptionFilters')

    def wait_for_log_events(self, log_group_name, log_stream_name, max_attempts=20 ,sleep_for=0.1):
        for i in range(0, max_attempts):
            result = self.log_events_raw(log_group_name=log_group_name, log_stream_name=log_stream_name)
            events = result.get('events')
            if len(events) > 0:
                #print(f' got logs after {i} * {sleep_for} seconds')        # note: when coding this this value was between 4 and 8
                return events
            wait_for(sleep_for)
        return []





import  boto3
from    time                                import sleep
from osbot_aws.apis.Session import Session
from osbot_aws.aws.iam.IAM import IAM
from osbot_utils.utils import Misc
from osbot_utils.utils.Dev import Dev


class CodeBuild:

    def __init__(self, project_name, role_name):
        self.codebuild    = Session().client('codebuild')
        self.iam          = IAM(role_name=role_name)
        self.project_name = project_name
        return

    def _invoke_via_paginator(self, method, field_id, use_paginator, **kwargs):
        paginator = self.codebuild.get_paginator(method)
        for page in paginator.paginate(**kwargs):
            for id in page.get(field_id):
                yield id
            if use_paginator is False:
                return

    def all_builds_ids(self, use_paginator = False):
        return self._invoke_via_paginator('list_builds','ids',use_paginator)

    def build_info(self, build_id):
        builds = self.codebuild.batch_get_builds(ids=[build_id]).get('builds')
        return Misc.array_pop(builds,0)

    def build_start(self):
        kvargs = { 'projectName': self.project_name }
        return self.codebuild.start_build(**kvargs).get('build').get('arn')

    def build_wait_for_completion(self, build_id, sleep_for=0.5, max_attempts=20, log_status=False):
        for i in range(0,max_attempts):
            build_info    = self.build_info(build_id)
            build_status  = build_info.get('buildStatus')
            current_phase = build_info.get('currentPhase')
            if log_status:
                Dev.pprint("[{0}] {1} {2}".format(i,build_status,current_phase))
            if build_status != 'IN_PROGRESS':
                return build_info
            sleep(sleep_for)
        return None


    def policies_create(self, policies):
        return self.iam.policies_create(policies, self.project_name)

    def project_builds(self,ids):
        return self.codebuild.batch_get_builds(ids=ids)

    def project_create(self, project_repo, service_role):

        kvargs = {
            'name': self.project_name,
            'source': {'type': 'GITHUB',
                       'location': project_repo},
            'artifacts': {'type': 'NO_ARTIFACTS'},
            'environment': {'type': 'LINUX_CONTAINER',
                            'image': 'aws/codebuild/python:3.7.1-1.7.0',
                            'computeType': 'BUILD_GENERAL1_SMALL'},
            'serviceRole': service_role
        }

        return self.codebuild.create_project(**kvargs)

    def project_delete(self):
        if self.project_exists() is False: return False
        self.codebuild.delete_project(name=self.project_name)
        return self.project_exists() is False

    def project_exists(self):
        return self.project_name in self.projects()


    def project_info(self):
        projects = Misc.get_value(self.codebuild.batch_get_projects(names=[self.project_name]),'projects',[])
        return Misc.array_pop(projects,0)

    def project_builds_ids(self, project_name, use_paginator=False):
        if use_paginator:
            kwargs = { 'projectName' : project_name }
        else:
            kwargs = { 'projectName' : project_name ,
                       'sortOrder'   : 'DESCENDING'  }
        return self._invoke_via_paginator('list_builds_for_project', 'ids',use_paginator, **kwargs)


    def projects(self):
        return self.codebuild.list_projects().get('projects')



from dotenv import load_dotenv

from osbot_aws.aws.bedrock.cache.Bedrock__Cache import Bedrock__Cache
from osbot_aws.aws.bedrock.Bedrock import Bedrock
from osbot_aws.aws.iam.IAM_Assume_Role import IAM_Assume_Role
from osbot_utils.decorators.lists.group_by import group_by
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.cache_on_self import cache_on_self


class Bedrock__with_temp_role(Bedrock):

    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.bedrock_cache = Bedrock__Cache()

    def iam_assume_role(self, service):
        load_dotenv()
        #policies_to_add = [dict(service=service, action=action, resource=resource)]
        services        = ["bedrock", "bedrock-runtime"]
        action          = "*"
        resource        = "*"
        role_name       = 'osbot__temp_role_for__test_Bedrock'
        policies_to_add = []
        for service in services:
            policies_to_add.append(dict(service=service, action=action, resource=resource))
        iam_assume_role = IAM_Assume_Role(role_name=role_name, policies_to_add=policies_to_add)
        iam_assume_role.create_role(recreate=False)
        #iam_assume_role.credentials_reset()
        return iam_assume_role

    @cache_on_self
    def client(self):
        service = "bedrock"
        return self.iam_assume_role(service).boto3_client(service_name=service, region_name=self.region_name)

    @cache_on_self
    def runtime(self):
        service = "bedrock-runtime"
        return self.iam_assume_role(service).boto3_client(service_name=service, region_name=self.region_name)


    # cached methods
    def model_invoke(self, model_id, body):
        return self.bedrock_cache.model_invoke(super(), model_id, body)

    def model_invoke_stream(self,model_id, body):
        return self.bedrock_cache.model_invoke_stream(super(), model_id, body)

    @index_by
    @group_by
    def models(self, **kwargs):
        return self.bedrock_cache.models(super(), **kwargs)




from osbot_utils.utils.Json import json_parse
from osbot_utils.utils.Misc import split_lines
from osbot_aws.helpers.Lambda_SSH import Lambda_SSH


class Lambda_SSH_Kubectl(Lambda_SSH):

    def kubectl(self, params):
        ssh_command = f"sudo kubectl {params}"
        return self.invoke_lambda(ssh_command)

    def deployments(self):
        return json_parse(self.kubectl('get deployments -A -o json'))

    def events(self):
        return json_parse(self.kubectl('get events -A -o json'))

    def nodes(self):
        return json_parse(self.kubectl('get nodes -A -o json'))

    def persistent_volumes(self):
        return json_parse(self.kubectl('get pv -A -o json'))

    def pod_logs(self, pod_namespace, pod_name, container_name):
        params = f'logs -n {pod_namespace} {pod_name} {container_name}'
        return split_lines(self.kubectl(params))

    def pod_exec(self, pod_namespace, pod_name, container_name, exec_command):
        params = f'exec -n {pod_namespace} {pod_name} {container_name} -- {exec_command}'
        return self.kubectl(params)

    def pods(self):
        return json_parse(self.kubectl('get pods -A -o json'))

    def services(self):
        return json_parse(self.kubectl('get services -A -o json'))



from osbot_aws.aws.ec2.EC2 import EC2
from osbot_aws.aws.ec2.EC2_Instance import EC2_Instance
from osbot_aws.aws.ec2.utils.EC2__Create__Instance import EC2__Create__Instance
from osbot_aws.utils.Version import Version
from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Misc import random_id, random_uuid_short

# todo , add methods to wait for the ami to be created and then stop the instance
class EC2__Create__AMI(Type_Safe):
    ec2_create_instance : EC2__Create__Instance
    ec2_instance        : EC2_Instance          = None
    instance_id         : str
    ami_name            : str
    ec2                 : EC2

    def create_instance(self):
        self.instance_id  = self.ec2_create_instance.create_instance__with_amazon_linux__t3_nano__ssh_sg__spot()
        self.ec2_instance = EC2_Instance(instance_id=self.instance_id, ec2=self.ec2)
        return self

    def create_ami_from_instance_id(self):
        instance_id = self.instance_id
        if not self.ami_name:
            self.ami_name = f'osbot-ami-{self.instance_id}-{random_uuid_short()}'
        return self.ec2_create_instance.ec2.create_image(instance_id=instance_id, name=self.ami_name)

    def ssh(self):
        return self.ec2_instance.ssh(self.ec2_create_instance.path_key_file(),'ec2-user' )

    def install_python_and_osbot_aws(self):
        self.ec2_instance.wait_for_ssh()
        with self.ssh() as _:
            _.ssh_execute().print_after_exec = True
            _.ssh_linux_amazon().install_python3()
            _.ssh_linux_amazon().pip_install('osbot_aws')

    def install_fastapi(self):
        with self.ssh() as _:
            _.ssh_linux_amazon().pip_install('fastapi')



    def create_ami__for__osbot_aws(self):
        region_name         = self.ec2_create_instance.ec2.aws_config.region_name()
        osbot_version       = Version().value()
        self.ami_name      = f'osbot-aws__{region_name}__{osbot_version}'
        self.create_instance()
        self.install_python_and_osbot_aws()
        self.create_ami_from_instance_id()

    def create_ami__for__osbot_fast_api(self):
        region_name = self.ec2_create_instance.ec2.aws_config.region_name()
        osbot_version = Version().value()
        self.ami_name = f'osbot-fastapi___{osbot_version}___{region_name}'
        self.create_instance()
        self.install_python_and_osbot_aws()
        self.install_fastapi()
        self.create_ami_from_instance_id()




import boto3
import requests
from botocore.config                                import Config

from osbot_aws.aws.s3.S3 import S3
from osbot_utils.base_classes.Type_Safe             import Type_Safe
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Env                          import get_env
from osbot_utils.utils.Status                       import status_ok, status_error

ENV_NAME__MINIO__SERVER     = 'MINIO__SERVER'
ENV_NAME__MINIO__ACCESS_KEY = 'MINIO__ACCESS_KEY'
ENV_NAME__MINIO__SECRET_KEY = 'MINIO__SECRET_KEY'

DEFAULT__MINIO__SERVER     = 'http://localhost:9000'
DEFAULT__MINIO__ACCESS_KEY = 'minioadmin'                       # todo: change this to a random value
DEFAULT__MINIO__SECRET_KEY = 'minioadmin'                       #       (from env vars, that are also used by minio docker image)

class S3__Minio(Type_Safe):

    def connect_to_server(self):
        try:
            response = requests.get(self.url_server())
            # Check if the status code is 200, indicating MinIO is up and running
            if response.status_code == 200:
                return status_ok(message="MinIO console is accessible on port 9001")
            return status_error(message="Server returned status code {response.status_code}")
        except requests.exceptions.ConnectionError as e:
            return status_error(message=f"Could not connect to MinIO console on port 9001: {e}")

    def connection_details(self):
        server     = get_env(ENV_NAME__MINIO__SERVER    , DEFAULT__MINIO__SERVER    )
        access_key = get_env(ENV_NAME__MINIO__ACCESS_KEY, DEFAULT__MINIO__ACCESS_KEY)
        secret_key = get_env(ENV_NAME__MINIO__SECRET_KEY, DEFAULT__MINIO__SECRET_KEY)
        return server, access_key, secret_key

    def s3(self):
        server, aws_access_key_id, aws_secret_access_key = self.connection_details()

        kwargs = dict(endpoint_url          = server                ,
                      aws_access_key_id     = aws_access_key_id     ,
                      aws_secret_access_key = aws_secret_access_key )
        s3 = S3(**kwargs)
        return s3

    def s3_client(self):
        server, access_key, secret_key = self.connection_details()
        minio_endpoint   = server
        minio_access_key = access_key
        minio_secret_key = secret_key

        # Create a session and S3 client configured to use MinIO
        s3_client = boto3.client('s3',
            endpoint_url          = minio_endpoint                  ,
            aws_access_key_id     = minio_access_key                ,
            aws_secret_access_key = minio_secret_key                ,
            config                = Config(signature_version='s3v4'),
            region_name           = 'us-east-1'                     )
        return s3_client

    @cache_on_self
    def server_online(self):
        return self.connect_to_server().get('status') == 'ok'

    def url_server(self):
        return 'http://localhost:9001'



from osbot_aws.apis.Session                             import Session
from osbot_utils.decorators.lists.group_by              import group_by
from osbot_utils.decorators.lists.index_by              import index_by
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value

class API_Gateway_V2:

    def __enter__(self                           ): return self
    def __exit__ (self, exc_type, exc_val, exc_tb): pass


    @cache_on_self
    def client(self):
        return Session().client('apigatewayv2')

    @remove_return_value(field_name='ResponseMetadata')
    def api(self, api_id):
        return self.client().get_api(ApiId=api_id)

    @group_by
    @index_by
    def apis(self):
        return self.client().get_apis().get('Items')

    def apis_by_name(self):
        return self.apis(index_by='Name')



from datetime import datetime, timedelta

import boto3
from osbot_utils.utils.Lists import unique
from osbot_aws.apis.Boto_Helpers import Boto_Helpers
from osbot_utils.utils.Files import file_create_from_bytes
from osbot_utils.utils.Json import json_parse, json_to_str
from osbot_utils.utils.Misc import list_set,wait
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.cache  import cache
from osbot_aws.apis.Session                import Session


class Cloud_Watch():
    def __init__(self):
        pass

    @cache
    def client(self):
        return Session().client('cloudwatch')

    def alarms(self):
        result = self.client().describe_alarms()
        return { "composite_alarms": result.get('CompositeAlarms'),
                 "metric_alarms"   : result.get('MetricAlarms'   )}

    def alarms_for_metric(self, metric_name, namespace):
        result = self.client().describe_alarms_for_metric(MetricName=metric_name, Namespace=namespace)
        return result

    def anomaly_detectors(self):
        result = self.client().describe_anomaly_detectors().get('AnomalyDetectors')
        return result

    def dashboard(self, dashboard_name):
        try:
            dashboard_data = self.client().get_dashboard(DashboardName=dashboard_name)
            dashboard = {"arn"    : dashboard_data.get('DashboardArn' ),
                         "widgets": (json_parse(dashboard_data.get('DashboardBody')).get('widgets')),
                         "name"   : dashboard_data.get('DashboardName') }

            return dashboard
        except Exception:          # can't use botocore.errorfactory.ResourceNotFound because that is not exposed
            return {}

    @index_by
    def dashboards(self):
        return self.client().list_dashboards().get('DashboardEntries')

    def dashboards_names(self):
        return list_set(self.dashboards(index_by='DashboardName'))

    def insight_rules(self):
        result = self.client().describe_insight_rules().get('InsightRules')
        return result

    def metric_add(self, namespace, metric_name, dimensions, value, storage_resolution=1, unit='None'):
        dimensions_data = []
        kwargs =  {"metric_data" : [ { 'MetricName'         : metric_name        ,
                                      'Dimensions'          : dimensions_data    ,
                                      'Unit'                : unit               ,
                                       "StorageResolution"  : storage_resolution ,
                                      'Value'               : value             }],
                    "namespace" : namespace}
        if dimensions:
            for name, value in dimensions.items():
                dimensions_data.append({'Name':name, "Value": value})
        return self.metric_add_raw(**kwargs)

    def metric_add_raw(self, namespace, metric_data):
        result = self.client().put_metric_data(Namespace=namespace, MetricData=metric_data)
        return result.get('ResponseMetadata',{}).get('HTTPStatusCode') == 200

    # see https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/CloudWatch-Metric-Widget-Structure.html#CloudWatch-Metric-Widget-Metrics-Array-Format
    # formula is [Namespace, MetricName, Dimension1Name, Dimension1Value, Dimension2Name, Dimension2Value...{Options Object}]
    def metric_widget_image(self, metric_widget, save_to_disk=True, path_image_file=None):
        if type(metric_widget) is not str:
            metric_widget = json_to_str(metric_widget)
        response = self.client().get_metric_widget_image(MetricWidget=metric_widget)
        png_bytes = response.get('MetricWidgetImage')
        if save_to_disk:
            return file_create_from_bytes(bytes=png_bytes, extension=".png", path=path_image_file)
        return png_bytes

    def metric_data(self, namespace, metric_name, dimensions=None, query_id='metric_query', start_time=None, end_time=None, period=60, stat='Average', unit=None):
        dimensions_data     = []
        metric_stat         = { 'Namespace' : namespace       ,
                                'MetricName': metric_name     ,
                                'Dimensions': dimensions_data }

        metric_data_query   = { 'Id'        : query_id ,
                                'MetricStat': { 'Metric' : metric_stat   ,
                                                'Period' : period        ,
                                                'Stat'   : stat          ,
                                                #'Unit'   : unit
                                                }}

        if unit:
            metric_data_query['MetricStat']['Unit'] = unit
        if dimensions:
            for name, value in dimensions.items():
                dimensions_data.append({"Name": name, "Value": value })


        if start_time is None:
            start_time= datetime.utcnow() - timedelta(minutes=60)

        if end_time is None:
            end_time=  datetime.utcnow()

        kwargs = { "MetricDataQueries": [metric_data_query] ,
                   "StartTime"        : start_time          ,
                   "EndTime"          : end_time            }

        return self.client().get_metric_data(**kwargs).get('MetricDataResults')

    def metric_data__wait_for_value(self, value, max_attempts=20, **metric_data_kwargs):
        for i in range(0,max_attempts):
            for result in self.metric_data(**metric_data_kwargs):   # this call takes about ~0.14 secs (so no need to add extra delay)
                if value in result.get("Values"):
                    return True
        return False




    def metric_image(self, namespace, metric_name, dimensions=None, options=None,
                           region=None, title=None, period=None, path_image_file=None,
                           start='-PT10M', end='P0D', stacked=True, height=400, width=600,
                           stat="Average", theme='dark', view='timeSeries'):
        if options is None:
            options = {}
        metric = [namespace, metric_name]
        if dimensions:
            for name, value in dimensions.items():
                metric.append(name )         # dimension_name_i
                metric.append(value)         # dimension_value_i
        if options:
            metric.append(options)
        metrics = [metric]

        return self.metrics_image(metrics,region=region, title=title, period=period, path_image_file=path_image_file,
                           start=start, end=end, stacked=stacked, height=height, width=width,
                           stat=stat, theme=theme, view=view)

    def metrics_image(self, metrics, region=None, title=None, period=None, path_image_file=None,
                           start='-PT10M', end='P0D', stacked=False, height=400, width=600,
                           stat="Average", theme='dark', view='timeSeries'):
        metric_widget = { "metrics": metrics}
        if region   : metric_widget['region' ] = region
        if title    : metric_widget['title'  ] = title
        if period   : metric_widget['period' ] = period
        if start    : metric_widget['start'  ] = start
        if end      : metric_widget['end'    ] = end
        if stacked  : metric_widget['stacked'] = stacked
        if width    : metric_widget['width'  ] = width
        if height   : metric_widget['height' ] = height
        if stat     : metric_widget['stat'   ] = stat   # SampleCount | Average | Sum | Minimum | Maximum | p?? | TM(??:??), TC(??:??) | TS(??:??) | WM(??:??) | PR(??:??) | IQM # see https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/Statistics-definitions.html
        if theme    : metric_widget['theme'  ] = theme  # light | dark
        if view     : metric_widget['view'   ] = view   # timeSeries | bar | pie

        return self.metric_widget_image(metric_widget=metric_widget, path_image_file=path_image_file)

    def metric_list_raw(self, namespace, metric_name=None, dimensions=None, recently_active='PT3H'):
        kwargs = {"Namespace": namespace}
        if metric_name:
            kwargs['MetricName'] = metric_name
        if dimensions:
            dimensions_data = []
            for key, value in dimensions.items():
                dimensions_data.append({"Key": key, "Value": value })
            kwargs['Dimensions'] = dimensions_data
        if recently_active:
             kwargs['RecentlyActive'] = recently_active
        return Boto_Helpers.invoke_using_paginator(self.client(), "list_metrics","Metrics", **kwargs)

    # this is paginated, so be careful with using recently_active=None and max_results not being set
    def metric_list(self, namespace, metric_name=None, dimensions=None, recently_active='PT3H', max_results=-1):
        results_paginated = self.metric_list_raw(namespace=namespace, metric_name=metric_name, dimensions=dimensions,recently_active=recently_active)
        results           = {}
        items_added       = 0
        for result in results_paginated:
            _dimensions  = result.get('Dimensions')
            _metric_name = result.get('MetricName')
            _namespace   = result.get('Namespace')

            if results.get(_namespace) is None:
                results[_namespace] = {}
            result_namespace = results[_namespace]

            if result_namespace.get(_metric_name) is None:
                result_namespace[_metric_name] = {}
            result_metric_name = result_namespace[_metric_name]

            for result_dimension in _dimensions:
                dimension_name  = result_dimension.get('Name')
                dimension_value = result_dimension.get('Value')

                if result_metric_name.get(dimension_name) is None:
                    result_metric_name[dimension_name] = []
                result_dimension_name = result_metric_name[dimension_name]

                result_dimension_name.append(dimension_value)
                items_added+=1
            if max_results !=-1 and items_added > max_results:
                break

        # sort and unique dimensions data
        for namespace in results.values():
            for metric_name, metric_values in namespace.items():
                for dimension_name, dimension_values in metric_values.items():
                    metric_values[dimension_name] = unique(metric_values[dimension_name])

        return results

    def metric_statistics(self, namespace, metric_name, dimensions=None, start_time=None, end_time=None, period=60, statistics=None, extended_statistics=None, unit=None):
        kwargs = {  "Namespace" : namespace   ,
                    "MetricName": metric_name ,
                    "Period"    : period
                }
        if dimensions:
            dimensions_data = []
            for name, value in dimensions.items():
                dimensions_data.append({"Name": name, "Value": value })
            kwargs['Dimensions'] = dimensions_data
        if start_time:
            kwargs['StartTime'] = start_time
        if start_time:
            kwargs['EndTime'] = end_time
        if statistics:
            kwargs['Statistics'] = statistics       # 'SampleCount'|'Average'|'Sum'|'Minimum'|'Maximum',
        if extended_statistics:
            kwargs['ExtendedStatistics'] = extended_statistics
        if unit:
            kwargs['Unit'] = unit                   # 'Seconds'|'Microseconds'|'Milliseconds'|'Bytes'|'Kilobytes'|'Megabytes'|'Gigabytes'|'Terabytes'|'Bits'|'Kilobits'|'Megabits'|'Gigabits'|'Terabits'|'Percent'|'Count'|'Bytes/Second'|'Kilobytes/Second'|'Megabytes/Second'|'Gigabytes/Second'|'Terabytes/Second'|'Bits/Second'|'Kilobits/Second'|'Megabits/Second'|'Gigabits/Second'|'Terabits/Second'|'Count/Second'|'None'
        return self.client().get_metric_statistics(**kwargs).get("Datapoints")




    # def metrics(self):                                          # todo add filter and pagination
    #     return self.client().list_metrics().get('Metrics')

    def namespaces_list(self):
        return ["AWS/AmplifyHosting" , "AWS/ApiGateway" , "AWS/AppRunner" , "AWS/AppStream" , "AWS/AppSync" ,
                "AWS/Athena" , "Backup" , "AWS/Billing" , "AWS/CertificateManager" , "AWS/ACMPrivateCA" ,
                "AWS/Chatbot" , "AWS/ClientVPN" , "AWS/CloudFront" , "AWS/CloudHSM" , "AWS/CloudSearch" ,
                "AWS/Logs" , "AWS/CodeBuild" , "AWS/CodeGuruProfiler" , "AWS/Cognito" , "AWS/Connect" ,
                "AWS/DataLifecycleManager" , "AWS/DataSync" , "AWS/DMS" , "AWS/DX" , "AWS/DocDB" , "AWS/DynamoDB" ,
                "AWS/DAX" , "AWS/EC2" , "AWS/ElasticGPUs" , "AWS/EC2Spot" , "AWS/AutoScaling" , "AWS/ElasticBeanstalk" ,
                "AWS/EBS" , "AWS/ECS" , "AWS/ECS/ManagedScaling" , "AWS/EFS" , "AWS/ElasticInference" , "AWS/ApplicationELB" ,
                "AWS/NetworkELB" , "AWS/GatewayELB" , "AWS/ELB" , "AWS/ElasticTranscoder" , "AWS/ElastiCache" ,
                "AWS/ElastiCache" , "AWS/ES" , "AWS/ElasticMapReduce" , "AWS/MediaConnect" , "AWS/MediaConvert" ,
                "AWS/MediaLive" , "AWS/MediaPackage" , "AWS/MediaStore" , "AWS/MediaTailor" , "AWS/Events" , "AWS/FSx" ,
                "AWS/FSx" , "AWS/GameLift" , "AWS/GlobalAccelerator" , "Glue" , "AWS/GroundStation" , "AWS/HealthLake" ,
                "AWS/Inspector" , "AWS/IVS" , "AWS/IoT" , "AWS/IoTAnalytics" , "AWS/IoTSiteWise" , "AWS/ThingsGraph" ,
                "AWS/KMS" , "AWS/Cassandra" , "AWS/KinesisAnalytics" , "AWS/Firehose" , "AWS/Kinesis" , "AWS/KinesisVideo" ,
                "AWS/Lambda" , "AWS/Lex" , "AWS/Location" , "AWS/LookoutMetrics" , "AWS/ML" , "AWS/Kafka" , "AWS/AmazonMQ" ,
                "AWS/Neptune" , "AWS/NetworkFirewall" , "AWS/OpsWorks" , "AWS/Polly" , "AWS/QLDB" , "AWS/Redshift" ,
                "AWS/RDS" , "AWS/Robomaker" , "AWS/Route53" , "AWS/SageMaker" , "AWS/SDKMetrics" , "AWS/ServiceCatalog" ,
                "AWS/DDoSProtection" , "AWS/SES" , "AWS/SNS" , "AWS/SQS" , "AWS/S3" , "AWS/SWF" , "AWS/States" ,
                "AWS/StorageGateway" , "AWS/SSM-RunCommand" , "AWS/Textract" , "AWS/Timestream" , "AWS/Transfer" ,
                "AWS/Translate" , "AWS/TrustedAdvisor" , "AWS/NATGateway" , "AWS/TransitGateway" , "AWS/VPN" , "AWS/WAFV2" ,
                "WAF" , "AWS/WorkMail" , "AWS/WorkSpaces"]

    def namespace_metrics(self, namespace, metric_name=None, dimensions=None, recently_active='PT3H'):
        result = self.metric_list(namespace=namespace, metric_name=metric_name, dimensions=dimensions, recently_active=recently_active)
        namespace_data = result.get(namespace,{})
        metrics = {}
        for metric_name, metric_data in namespace_data.items():
            metrics[metric_name] = list_set(metric_data)
        return metrics

    def tags_for_resource(self, resource_arn):
        result = self.client().list_tags_for_resource(ResourceARN=resource_arn)
        return result




from osbot_aws.aws.ecs.ECS_Fargate_Task import ECS_Fargate_Task


class Temp_ECS_Fargate_Task:
    def __init__(self, image_name, delete_on_exit=False):
        cluster_name        = f'Temp_ECS_Fargate_Task-{image_name}'
        self.delete_on_exit = delete_on_exit
        self.fargate_task   = ECS_Fargate_Task(cluster_name=cluster_name, image_name=image_name)

    def __enter__(self):
        self.fargate_task.setup()
        return self.fargate_task

    def __exit__(self, exception_type, exception_value, traceback):
        if self.delete_on_exit:
            self.fargate_task.delete_cluster_task_definition_tasks_roles()


from osbot_aws.aws.iam.IAM_Role import IAM_Role
from osbot_aws.aws.iam.roles.IAM_Policy__Full_Access import IAM__Policy__Service__Full_Access
from osbot_aws.aws.iam.roles.IAM__Policy__Assume__Service import iam_statement__ec2__assume_service
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Dev import pprint


class IAM__Role_For(Kwargs_To_Self):
    assume_policy : dict
    iam_role      : IAM_Role = None
    skip_if_exists: bool     = True

    def create(self):
        kwargs = dict(assume_policy_document = {'Statement': [self.assume_policy] }  ,
                      skip_if_exists         = self.skip_if_exists )
        return self.iam_role.create(**kwargs)

    def exists(self):
        return self.iam_role.exists()

    def role_policies(self):
        return list(self.iam_role.iam.role_policies_attached())

class IAM__Role_for__EC2_Instances(IAM__Role_For):

    def __init__(self, role_name):
        self.assume_policy = iam_statement__ec2__assume_service()
        self.iam_role = IAM_Role(role_name)
        super().__init__()

    def attach_policy__full_access__s3(self):
        iam_policy = IAM__Policy__Service__Full_Access(service_name='s3')
        iam_policy.create()                                                 # create if doesn't exist
        iam_policy_arn = iam_policy.arn()
        return self.iam_role.iam.role_policy_attach(policy_arn=iam_policy_arn)

    def setup(self):
        if self.exists() is False:
            create_result = self.create()
            assert create_result.get('status') == 'ok'
            self.attach_policy__full_access__s3()
            self.iam_role.create_instance_profile()         # this is needed so that EC2 can use this role
            self.iam_role.add_to_instance_profile()
        return self




from osbot_aws.aws.sts.STS import STS
from osbot_aws.deploy.Deploy_Lambda import Deploy_Lambda

from osbot_utils.utils.Files import file_name, file_contents

from osbot_aws.lambdas.server_ssh import run
from osbot_aws.apis.Secrets import Secrets


class Lambda_SSH:

    def __init__(self, key_name, target_host):
        self.sts            = STS()
        self.region         = self.sts.current_region_name()
        self.key_name       = key_name
        self.target_host    = target_host
        self.lambda_handler = run
        self.lambda_layer   = f"arn:aws:lambda:{self.region}:553035198032:layer:git-lambda2:8"      # todo: rebuild this layer (https://github.com/lambci/git-lambda-layer)
        self.ssh_user       = 'ubuntu'
        self.deploy         = Deploy_Lambda(self.lambda_handler)
        self.lambda_name    = self.deploy.package.lambda_name
        self.aws_lambda     = self.deploy.package.aws_lambda

    def deploy_lambda(self):
        self.deploy.add_osbot_aws()
        if self.deploy.deploy():
            self.aws_lambda.set_layers([self.lambda_layer])
            self.aws_lambda.update_lambda_configuration()
        return self.aws_lambda.exists()

    def invoke_lambda(self, ssh_command):
        ssh_key    = self.get_key_from_secret_store()
        payload = { 'target_host' : self.target_host,
                    'ssh_key'     : ssh_key         ,
                    'ssh_key_name': self.key_name   ,
                    'ssh_user'    : self.ssh_user   ,
                    'ssh_command' : ssh_command     }

        return self.aws_lambda.invoke(payload)

    def get_key_from_secret_store(self):
        secret_name = f"ssh_key_{self.key_name}.pem"
        return Secrets(secret_name).value()

    def upload_ssh_key_to_secret_store(self, path_to_ssh_key):
        key_name    = file_name(path_to_ssh_key)
        secret_name = f"ssh_key_{key_name}"
        secret_data = file_contents(path_to_ssh_key)
        return Secrets(secret_name).create(secret_data)